!
! Module: region_mod
!
! Authors:  Richard West - SSC (Leicester)
!	    Ian Stewart (SSC/LUX)
!          
! Description: Create source/background region definition files for 
! extraction of products
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module region_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  use cal
  use param
  use errorHandling
  use datafilter

!  use ascregion_mod	! in mssllib. Contains subs to write regions.
use ims_ascregion_mod

  use array_utils	! for smallest/largestTrueIndex
  use cal_aux
  use confusion
  use constants
  use coordinate
  use dal_aux	! for readColToFixed
  use ellipse	! for EllipseT
  use psf_ims
  use source_cutouts	! for SrcDataT
  use ssc_misc	! for intToStr
  use wcs_aux

  implicit none 

  private 

  ! Command-line parameters:

  character(255) ::&
    evlistSetName = '',&
    srcListSetTabName = '',&
    tempSrcSetName = '',&
    selectionExpression = '',&
    srcRegionSetName = '',&
    bkgRegionSetName = '',&
    bkgKwdSetName = '',&
    bkgImgSetName = ''
  character(20) ::&
    outUnit = '',&
    opMode = '',&
    radiusMode = '',&
    srcIdColName = '',&
    bkgFluxStyle = '',&
    bkgRateColName = ''
  integer ::&
    userSrcNum = 0
  real(single) ::&
    userBkgFluxPerPixel = 0.0,&
    fracEncircEnergy = 0.0,&
    fracBackground = 0.0,&
    userSrcRadiusArcsec = 0.0
  logical(bool) ::&
    doEllipse = .false.,&
    annulusConcentricAboutFovCtr = .false.,&
    retainEllipticalBkgAnnuli = .false.,&
    adjustRadius = .false.

  ! Other globals:

  integer(int32) ::&
    instrumentId
  real(double) ::&
    expStartTimeStamp
  type(SpacecraftAttitudeType) ::&
    scAttitude, instrumAttitude
  type(WcsT) ::&
    wcs

  type :: localAnnulusType
    real(double) ::&
      centreX,&
      centreY
    real(single) ::&
      innerLongAxis,&
      outerLongAxis
    type(EllipseT) ::& ! typedef in ssclib/ellipse
      ellipse
    character(10) :: unit	! eg 'pixels', 'detxy', 'tanxy'
    logical(bool) :: isAnnulus	! FALSE means the inner area is of zero size.
    ! This should override any non-zero value in innerLongAxis.
  end type localAnnulusType

  type :: localRegListType
    type(ims_RegionT), pointer :: regions(:)
  end type localRegListType

  public region

contains 
  
  !---------------------------------------------------------------------
  subroutine region
  !---------------------------------------------------------------------

    implicit none
    type(DataSetT) :: set, sourceSet
    type(TableT) :: tab, sourceTab
    character(255) :: srcListSetName, srcListTabName
    integer :: numFilteredRows, srcNum, trialIdInst, trialIdBand
    integer, parameter :: maxSrcNum = 9999
    type(WcsT) :: srcWcs
    type(SrcDataT), allocatable :: srcData(:) ! typedef in ssclib/source_cutouts
    integer(int32), allocatable :: srcIdBand(:), srcIdInst(:)
    real(double) ::&
      bandLo(3,0:9), bandHi(3,0:9), meanEeV, userSrcRadiusTan,&
      smallestPixelDim, fovCtrTanX, fovCtrTanY, semiRadius,&
      opticAxisOffsetAngle, opticAxisDirectionAngle,&
      pixelSizeXWorld, pixelSizeYWorld
    logical(bool) :: infoFoundThisBand(3,0:9)
    real(double), allocatable ::&
      tanX(:), tanY(:), bkgAnnulusCtrTanX(:), bkgAnnulusCtrTanY(:),&
      srcTheta(:), nearestDistances(:), farthestDistances(:),&
      meanDistances(:), lengthsCentreRay(:)
    real(single), allocatable ::&
      srcShrinkage(:), dx(:), dy(:), meanRadii(:), xPixels(:), yPixels(:)
    type(localAnnulusType), allocatable :: bkgAnnulus(:)
    real(single) ::&
      fovCtrDetX, fovCtrDetY, fovCtrPixelsX, fovCtrPixelsY, diff, ave
    type(ims_RegionExtnT), allocatable :: regionExtnData(:)
    type(ims_RegionT) :: srcRegion
    type(ims_RegionT), pointer :: bkgRegions(:)
    type(localRegListType), pointer :: bkgRegList(:)
    type(ims_RegionT), allocatable :: srcRegions(:)
    character(4) :: srcIdStr
    character(20) :: xColName, yColName

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    select case(outUnit)
    case('xy')
      xColName = 'X'
      yColName = 'Y'

    case('detxy')
      xColName = 'DETX'
      yColName = 'DETY'

    case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
    end select

    !...................................................................
    ! Set up the cal and get wcs info if appropriate:

    call message(AppMsg, verbose, "Obtaining essential information from &
      &the event list.")

    set = dataSet(evlistSetName, READ)

    if (outUnit == 'xy') then
      tab = table(set, 'EVENTS')
      call readWcs(tab, 'X', 'Y', wcs) ! in ssclib/wcs_aux

      pixelSizeXWorld = sqrt(wcs%x%matrixOnAxis * wcs%x%matrixOnAxis&
       + wcs%x%matrixOffAxis * wcs%x%matrixOffAxis)
      pixelSizeYWorld = sqrt(wcs%y%matrixOnAxis * wcs%y%matrixOnAxis&
       + wcs%y%matrixOffAxis * wcs%y%matrixOffAxis)
!*** this is only valid if there is no skew.
    end if

    call getCalInfo(set, instrumentId, expStartTimeStamp, scAttitude&
      , instrumAttitude) ! in ssclib/cal_aux. Sets CAL instrument state.

    if (instrumentId .ne. EMOS1&
    .and. instrumentId .ne. EMOS2&
    .and. instrumentId .ne. EPN) then
call error('badInstrum', 'Instrument '//trim(instId(instrumentId))//&
&' not supported.')
    end if

    call release(set)

    !...................................................................
    ! Filter and load the source list:
!*** exactly the same as in imageutils/srcmask.

    call message(AppMsg, Verbose, 'Filtering source list.')

    call splitSetTabName(srcListSetTabName, srcListSetName, srcListTabName)
    ! in ssclib/dal_aux
    sourceSet = clone(srcListSetName, tempSrcSetName, TEMP)
    sourceTab = table(sourceSet, srcListTabName, MODIFY)

    call initFilter(selectionExpression)
    numFilteredRows = doFilter(sourceTab)

    if (numFilteredRows <= 0)&
call error('noSourcesSelected'&
, 'No sources passed the specified filtering expression.')

if (userSrcNum > numFilteredRows) call error('tooFewRows',"There &
&are fewer rows in the filtered source list than --usersrcid.")
!if (userSrcNum > maxSrcNum) call error('userSrcNumTooHigh',"The value &
!&of the --usersrcid parameter exceeds the maximum allowed.")

    if (numFilteredRows > maxSrcNum) then
      call warning('tooManyRows',"The number of rows in the filtered source &
        &list exceeds the maximum value of --usersrcid.")
    end if

    if (outUnit == 'xy' .and. radiusMode .eq. 'contour') then
      call readWcsSourceList(sourceSet, srcWcs)
!*** they should be stored as proper column keywords, but until Potsdam change their code there's not much I can do.
    end if

    allocate(srcData(numFilteredRows))
    call initializeSrcDataT(srcData) ! in ssclib/source_cutouts

    allocate(srcIdBand(numFilteredRows), srcIdInst(numFilteredRows))

    call readColDataToFixed(sourceTab, 'RA',         srcData%raDeg)
    call readColDataToFixed(sourceTab, 'DEC',        srcData%decDeg)
    call readColDataToFixed(sourceTab, 'SCTS',       srcData%flux)
    call readColDataToFixed(sourceTab, 'BG_MAP',     srcData%bkgFluxPerPixel)

    call readColDataToFixed(sourceTab, 'ID_INST',    srcIdInst)
!*** get status and so make the presence of this column optional.
    call readColDataToFixed(sourceTab, 'ID_BAND',    srcIdBand)
!*** get status and so make the presence of this column optional.

    if (opMode .eq. 'batch')&
      call readColDataToFixed(sourceTab, srcIdColName, srcData%idNum)
    ! The sub readColDataToFixed is in ssclib/dal_aux.

    call getBandEdges(sourceSet, bandLo, bandHi, infoFoundThisBand)

    call release(sourceSet)

    if (opMode == 'batch') then
      do srcNum = 1, numFilteredRows
        if (count(srcData%idNum == srcData(srcNum)%idNum) > 1)&
call error('nonUniqueSrcId',"The &
&source ID number is not unique in the filtered source set.")
      end do
    end if

    if (opMode .ne. 'single'&
    .and. count(srcIdBand == srcIdBand(1)) .ne. numFilteredRows) then
      call warning('nonUniqueIdBand',"More than one value of ID_BAND was &
        &detected. This implies that several concentric regions may be &
        &produced for a single source. If you don't want this to happen you &
        &should filter the source list (the task can do this for you via its &
        &--expression parameter) so that only one value of ID_BAND remains.")
    end if

    if (radiusMode .ne. 'userfixed') then
      !.................................................................
      ! Calculate source energies:

      call message(AppMsg, verbose, "Calculating a representative energy &
        &for each source.")

if (srcIdBand(1) < 0 .or. srcIdBand(1) > 9) call error('badIdBand',"The &
&value of the ID_BAND column of the source list must be between 0 and 9 &
&inclusive.")

      srcData%EeV = 500.0 ! default.
      do trialIdInst = 1, 3
        do trialIdBand = 0, 9
          if (count(srcIdInst == trialIdInst&
          .and. srcIdBand == trialIdBand)<=0)&
        cycle

          if (count(srcIdInst == trialIdInst.and.srcIdBand == trialIdBand) > 0&
          .and. .not. infoFoundThisBand(trialIdInst, trialIdBand)) then
            call warning('noBandEnergy',"No energy bounds were found for &
              &sources of this value of ID_INST and ID_BAND. The default value &
              &of mean source energy has been used instead.")
!*** put in mumbers.
        cycle
          end if

if (bandLo(trialIdInst, trialIdBand) >= bandHi(trialIdInst, trialIdBand))&
call error('badEnergyBandEdges',"The 'lower' energy is higher than the &
&'higher' energy for this band!")

          meanEeV = (bandLo(trialIdInst, trialIdBand)&
            + bandHi(trialIdInst, trialIdBand)) / 2.0

          where(srcIdInst == trialIdInst .and. srcIdBand == trialIdBand)&
            srcData%EeV = meanEeV
        end do
      end do
    end if

    !...................................................................
    ! Get the positions of the source centres in the appropriate coord
    ! system:

    call message(AppMsg, verbose, "Calculating the location of each source &
      &in the specified coordinate system.")

    ! There are combinations of command-line parameters for which it is
    ! unnecessary to calculate either DETXY or TELCOORD positions of the
    ! sources. However rather than set up a complicated 'if' statement
    ! which might then become incorrect if facilities were added or
    ! changed at a later date, I have decided that it would be simpler
    ! just to calculate both DETXY or TELCOORD positions under all
    ! circumstances. The time penalty involved is insignificant.

    call raDecToInst(srcData%raDeg, srcData%decDeg, scAttitude&
      , expStartTimeStamp, srcData%xDet, srcData%yDet, srcData%thetaArcsec&
      , srcData%phi) ! in ssclib/coordinate

    select case(outUnit)
    case('xy')
      if (radiusMode == 'userfixed' .or. radiusMode == 'enfrac') then

        allocate(tanX(numFilteredRows), tanY(numFilteredRows))
        call raDecToTan(wcs%x%refWorld, wcs%y%refWorld, srcData%raDeg&
          , srcData%decDeg, tanX, tanY) ! in ssclib/coordinate

        call tanToPixels(wcs, tanX, tanY, srcData%xPixel, srcData%yPixel)
        ! in ssclib/wcs_aux

        deallocate(tanX, tanY)
      end if
      ! Note that if radiusMode == 'contour' then source centre xy positions
      ! are obtained later as a by-product of calculating the hole locii.

    case('detxy')
!      if (radiusMode == 'userfixed') then
!        call raDecToInst(srcData%raDeg, srcData%decDeg, scAttitude&
!          , expStartTimeStamp, srcData%xDet, srcData%yDet)
!         ! in ssclib/coordinate
!      end if
!      ! Note that if radiusMode == 'contour' then source centre detxy positions
!      ! are obtained later as a by-product of calculating the hole locii;
!      ! whereas if radiusMode == 'enfrac', detxy positions have already been
!      ! obtained.

    case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
    end select

    !...................................................................
    ! Get the edges of the source holes:

    call message(AppMsg, verbose, "Calculating the source region locii.")

    select case(radiusMode)
    case('contour')
      select case(bkgFluxStyle)
      case('col') ! stick with what we have
      case('kwd')
!*** must divide by exposure?? Don't allow this mode yet.
call error('notSupported', "This mode is not yet supported.")
        set = dataSet(bkgKwdSetName, READ)
        srcData%bkgFluxPerPixel = real32Attribute(set, 'BKG_RATE')
        call release(set)

      case('image')
call error('notSupported', "This mode is not yet supported.")
      case('user')
        srcData%bkgFluxPerPixel = userBkgFluxPerPixel
      case default
call error('badBkgFluxStyle', "The value of --bkgratestyle not recognized.")
      end select

!*** temporarily:
if (outUnit == 'detxy') call error('notSupported',"--radiusstyle=contour, &
&--outunit=detxy is not yet supported.")

      call getSourceCutouts(wcs, srcWcs, scAttitude, expStartTimeStamp&
        , fracBackground, srcData)
      ! in ssclib/source_cutouts.
      !
      ! Note! This returns both xy and detxy coords of the source positions
      ! in the srcData structure.

!!!      if (outUnit == 'detxy') then
!*** transfer src region to detxy.
!!!      end if

      if (doEllipse) then
        srcData%region%isCircle = .false.
      else
        srcData%region%isCircle = .true.
        srcData%region%shortSemiAxis = srcData%region%longSemiAxis
      end if

    case('userfixed')
      select case(outUnit)
      case('xy')
        srcData%region%unit = 'pixels'
        userSrcRadiusTan = (userSrcRadiusArcsec / 3600.0) * PiOn180
        smallestPixelDim = min(pixelSizeXWorld, pixelSizeYWorld)

        srcData%region%longSemiAxis = userSrcRadiusTan / smallestPixelDim&
          / PiOn180

      case('detxy')
        srcData%region%unit = 'detxy'
        srcData%region%longSemiAxis = userSrcRadiusArcsec / detUnitArcsec
        ! detUnitArcsec is defined in ssclib/coordinate.

      case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
      end select

      srcData%region%isCircle = .true.
      srcData%region%shortSemiAxis = srcData%region%longSemiAxis

    case('enfrac')
      call getEncircEnergyRadii(srcData%thetaArcsec, srcData%phi&
        , srcData%EeV, fracEncircEnergy, srcData%region%longSemiAxis)
      ! returns the radius in tangent units.

      select case(outUnit)
      case('xy') ! a 'squashed' transform between inst and sky-image
      ! coordinates can distort the circles into ellipses but I'm going
      ! to ignore this for the moment and keep circles, whatever the image
      ! transform.

        srcData%region%unit = 'pixels'
        smallestPixelDim = min(pixelSizeXWorld, pixelSizeYWorld)

        srcData%region%longSemiAxis = srcData%region%longSemiAxis&
          / smallestPixelDim / PiOn180

      case('detxy')
        srcData%region%unit = 'detxy'
        srcData%region%longSemiAxis = (srcData%region%longSemiAxis * 3600.0&
          / PiOn180) / detUnitArcsec
        ! detUnitArcsec is defined in ssclib/coordinate.

      case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
      end select

      srcData%region%isCircle = .true.
      srcData%region%shortSemiAxis = srcData%region%longSemiAxis
    case default
call error('badRadiusStyle', "The value of --radiusstyle not recognized.")
    end select

    !.................................................................
    ! Shrink up confused regions if desired and possible:

    allocate(srcShrinkage(numFilteredRows))
    srcShrinkage = 1.0

    if ((radiusMode .ne. 'contour' .or. .not. doEllipse)&
    .and. opMode .ne. 'global') then

      call message(AppMsg, verbose, "Shrinking confused source regions.")

      select case(outUnit)
      case('xy')
        call correctRadii(srcData%xPixel, srcData%yPixel&
          , srcData%region%longSemiAxis, srcShrinkage)

      case('detxy')
        call correctRadii(srcData%xDet, srcData%yDet&
          , srcData%region%longSemiAxis, srcShrinkage)

      case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
      end select
    end if

    !...................................................................
    ! Construct the background annuli:

!**** do only if opMode ne global!

    call message(AppMsg, verbose, "Calculating the locii of the &
      &local-background regions.")

if (opMode .eq. 'single') then
  allocate(bkgAnnulus(1))
  allocate(bkgAnnulusCtrTanX(1))
  allocate(bkgAnnulusCtrTanY(1))
else
    allocate(bkgAnnulus(numFilteredRows))
    allocate(bkgAnnulusCtrTanX(numFilteredRows))
    allocate(bkgAnnulusCtrTanY(numFilteredRows))
end if

    call initializeAnnulusType(bkgAnnulus)

    if (annulusConcentricAboutFovCtr) then ! get FOV coords in both detxy
    ! and xy frames:
      call telCoordToDetXY(0.0d0, 0.0d0, fovCtrDetX, fovCtrDetY)
      ! in ssclib/coordinate

      if (outUnit == 'xy') then
        call instToTan(fovCtrTanX, fovCtrTanY, wcs%x%refWorld, wcs%y%refWorld&
          , scAttitude, expStartTimeStamp, thetaArcsec=0.0d0, phi=0.0d0)
        ! in ssclib/coordinate

        call tanToPixels(wcs, fovCtrTanX, fovCtrTanY, fovCtrPixelsX&
          , fovCtrPixelsY) ! in ssclib/wcs_aux
      end if
    end if

    select case(outUnit)
    case('xy')
      bkgAnnulus%unit = 'pixels'

      if (annulusConcentricAboutFovCtr) then
        if (retainEllipticalBkgAnnuli) then
          ! (1) Project a cone concentric about the optic axis onto the
          ! image plane. To do this:
          !   (a) first calculate the angle between the optic axis and the
          ! normal to the tangent plane:

          semiRadius = sqrt(fovCtrTanX * fovCtrTanX + fovCtrTanY * fovCtrTanY)

if (opMode .eq. 'single') then
  allocate(srcTheta(1))
  srcTheta(1) = srcData(userSrcNum)%thetaArcsec * PiOn180 / 3600.0
else
          allocate(srcTheta(numFilteredRows))
          srcTheta = srcData%thetaArcsec * PiOn180 / 3600.0
end if

          if (semiRadius > 0.0) then
            opticAxisOffsetAngle = atan(semiRadius)

            ! (b) Now calculate the distance along the tangent plane from
            ! its origin to both the nearest and furthest points of the
            ! projected ellipse. These points must necessarily lie upon
            ! the same vector from the origin of the tangent plane; the
            ! mid point between them will give the centre of the projected
            ! ellipse, and its long semiaxis will be equal to half the
            ! distance between them.

if (opMode .eq. 'single') then
  allocate(nearestDistances(1))
  allocate(farthestDistances(1))
  allocate(meanDistances(1))
  allocate(lengthsCentreRay(1))
else
            allocate(nearestDistances(numFilteredRows))
            allocate(farthestDistances(numFilteredRows))
            allocate(meanDistances(numFilteredRows))
            allocate(lengthsCentreRay(numFilteredRows))
end if

            nearestDistances = tan(opticAxisOffsetAngle - srcTheta) ! vector op.
!*** check that when the argument is negative < 90, tan returns negative!!

            farthestDistances= tan(opticAxisOffsetAngle + srcTheta) ! vector op.
            meanDistances = (nearestDistances + farthestDistances) / 2.0
            bkgAnnulus%ellipse%longSemiAxis = (farthestDistances&
              - nearestDistances) / 2.0

            ! (b) then find the angle on the tangent plane between the
            ! x-axis of this system and the location of the point at which
            ! the optic axis crosses the tangent plane;

            opticAxisDirectionAngle = atan2(fovCtrTanY, fovCtrTanX)
            bkgAnnulus%ellipse%rotatedAngle = opticAxisDirectionAngle + PiOn2

            ! To explain the added pi/2 (90 deg), consider the following.
            ! Suppose the opticAxisDirectionAngle was small - ie, the
            ! optic axis crossed the tangent plane close to the x-axis of
            ! the plane. The long axis of the projected ellipse would in
            ! this case be nearly aligned with the x-axis of the tangent
            ! plane. However the EllipseT structure element rotatedAngle
            ! is defined such that it is the angle (in the direction from
            ! positive x to positive y axis) through which the ellipse must
            ! be rotated to reach its present orientation, _starting with
            ! the short axis aligned with the x-axis of the coordinate
            ! sytem_. Thus in the example above in which the _long_ axis
            ! lies nearly along the x-axis of the plane, rotatedAngle must
            ! be a little over pi/2.

            bkgAnnulusCtrTanX = meanDistances * cos(opticAxisDirectionAngle)
            bkgAnnulusCtrTanY = meanDistances * sin(opticAxisDirectionAngle)
            lengthsCentreRay = sqrt(1.0 + meanDistances * meanDistances)
            bkgAnnulus%ellipse%shortSemiAxis = lengthsCentreRay&
              * tan(srcTheta)

            deallocate(nearestDistances, farthestDistances&
              , meanDistances, lengthsCentreRay)

          else ! semiRadius == 0.0. This means that the optic axis is
          ! normal to the tangent plane and the projection of the
          ! equal-theta cone will be a circle.

            bkgAnnulus%ellipse%longSemiAxis = tan(srcTheta)

            bkgAnnulus%ellipse%shortSemiAxis&
              = bkgAnnulus%ellipse%longSemiAxis

            bkgAnnulus%ellipse%rotatedAngle = 0.0
            bkgAnnulusCtrTanX = 0.0d0
            bkgAnnulusCtrTanY = 0.0d0
          end if

          deallocate(srcTheta)

          ! (2) Now rescale the ellipse from tan to pixel coordinates:

!          do srcNum = 1, numFilteredRows
do srcNum = 1, size(bkgAnnulus)
            call ellipseAngleToPhase(&
              bkgAnnulus(srcNum)%ellipse%shortSemiAxis,&
              bkgAnnulus(srcNum)%ellipse%longSemiAxis,&
              bkgAnnulus(srcNum)%ellipse%rotatedAngle,&
              bkgAnnulus(srcNum)%ellipse%xAmp,&
              bkgAnnulus(srcNum)%ellipse%yAmp,&
              bkgAnnulus(srcNum)%ellipse%phase&
            ) ! in ssclib/ellipse
          end do

if (opMode .eq. 'single') then
  allocate(xPixels(1), yPixels(1))
else
          allocate(xPixels(numFilteredRows), yPixels(numFilteredRows))
end if

          call tanToPixels(wcs, bkgAnnulusCtrTanX&
            , bkgAnnulusCtrTanY, xPixels, yPixels) ! in ssclib/wcs_aux

          bkgAnnulus%centreX = xPixels
          bkgAnnulus%centreY = yPixels
          deallocate(xPixels, yPixels)

!          bkgAnnulus%ellipse%xAmp = bkgAnnulus%ellipse%xAmp&
!            / abs(wcs%x%pixelSizeWorld) / PiOn180
!          bkgAnnulus%ellipse%yAmp = bkgAnnulus%ellipse%yAmp&
!            / abs(wcs%y%pixelSizeWorld) / PiOn180
bkgAnnulus%ellipse%xAmp = bkgAnnulus%ellipse%xAmp&
  / pixelSizeXWorld / PiOn180
bkgAnnulus%ellipse%yAmp = bkgAnnulus%ellipse%yAmp&
  / pixelSizeYWorld / PiOn180
!**** There are several things wrong with this. Easiest to fix or at least trap is that it assumes worldUnits are both 'deg' and that ctype = 'RA---TAN/DEC--TAN' etc. Harder to trap is that the above assumes there is no skew between pixel and world coords. Finally (altho its a related issue, if I found a way to handle skew it would also handle this) the parity of the transform gets lost in the way that pixelSizeXWorld are calculated. pixelSizeXWorld and pixelSizeYWorld end up as always positive.

!          do srcNum = 1, numFilteredRows
do srcNum = 1, size(bkgAnnulus)
            call ellipsePhaseToAngle(&
              bkgAnnulus(srcNum)%ellipse%xAmp,&
              bkgAnnulus(srcNum)%ellipse%yAmp,&
              bkgAnnulus(srcNum)%ellipse%phase,&
              bkgAnnulus(srcNum)%ellipse%shortSemiAxis,&
              bkgAnnulus(srcNum)%ellipse%longSemiAxis,&
              bkgAnnulus(srcNum)%ellipse%rotatedAngle&
            ) ! in ssclib/ellipse
          end do

          ! Now check how nearly circular they are and, if there is less
          ! than 1 part in 1000 difference, force them to be circular:

!          do srcNum = 1, numFilteredRows
do srcNum = 1, size(bkgAnnulus)
            diff = bkgAnnulus(srcNum)%ellipse%shortSemiAxis&
              - bkgAnnulus(srcNum)%ellipse%longSemiAxis
            ave = (bkgAnnulus(srcNum)%ellipse%shortSemiAxis&
              + bkgAnnulus(srcNum)%ellipse%longSemiAxis) / 2.0

            if (abs(diff) < ave / 1000.0) then
              bkgAnnulus(srcNum)%ellipse%isCircle = .true.
            end if
          end do

        else ! .not. retainEllipticalBkgAnnuli. Just make them circles
        ! in the tangent plane, although concentric about the optic axis
        ! still.

if (opMode .eq. 'single') then
  allocate(dx(1), dy(1))
  dx(1) = srcData(userSrcNum)%xPixel - fovCtrPixelsX
  dy(1) = srcData(userSrcNum)%yPixel - fovCtrPixelsY
else
          allocate(dx(numFilteredRows), dy(numFilteredRows))
          dx = srcData%xPixel - fovCtrPixelsX
          dy = srcData%yPixel - fovCtrPixelsY
end if
          bkgAnnulus%ellipse%longSemiAxis = sqrt(dx * dx + dy * dy)
          deallocate(dx, dy)

          bkgAnnulus%ellipse%shortSemiAxis&
            = bkgAnnulus%ellipse%longSemiAxis

          bkgAnnulus%ellipse%rotatedAngle = 0.0
          bkgAnnulus%centreX = fovCtrPixelsX
          bkgAnnulus%centreY = fovCtrPixelsY
          bkgAnnulus%ellipse%isCircle = .true.
        end if

        ! Now calculate the inner and outer radii. After these are
        ! calculated, normalize the short and long axes so that the long
        ! equals 1. The 'short' axis will then store the ratio between
        ! short and long (ie a number between 0 and 1). Recipes then will
        ! be:
        !
        !...............................................................
        ! Desired result     Recipe:
        !...............................................................
        !   SSA inner       innerLongAxis * ellipse%shortSemiAxis
        !   LSA inner       innerLongAxis
        !   SSA outer       outerLongAxis * ellipse%shortSemiAxis
        !   LSA outer       outerLongAxis
        !...............................................................
        !
        ! Note that the inner and outer radii are supposed to 'kiss' the
        ! inner and outer edge of the respective source region, but in the
        ! case of elliptical sources and/or annuli this condition is
        ! fulfilled only approximately, since to do it properly is not
        ! worth the extra algebra.

if (opMode .eq. 'single') then
  bkgAnnulus(1)%outerLongAxis = bkgAnnulus(1)%ellipse%longSemiAxis&
    + srcData(userSrcNum)%region%longSemiAxis
  bkgAnnulus(1)%innerLongAxis = bkgAnnulus(1)%ellipse%longSemiAxis&
    - srcData(userSrcNum)%region%longSemiAxis
else
        bkgAnnulus%outerLongAxis = bkgAnnulus%ellipse%longSemiAxis&
          + srcData%region%longSemiAxis
        bkgAnnulus%innerLongAxis = bkgAnnulus%ellipse%longSemiAxis&
          - srcData%region%longSemiAxis
end if

        where (bkgAnnulus%innerLongAxis <= 0.0)&
          bkgAnnulus%isAnnulus = .false.

        bkgAnnulus%ellipse%shortSemiAxis&
          = bkgAnnulus%ellipse%shortSemiAxis&
          / bkgAnnulus%ellipse%longSemiAxis

        bkgAnnulus%ellipse%longSemiAxis = 1.0

      else ! annulus concentric about src ctr:

if (opMode .eq. 'single') then
  bkgAnnulus(1)%centreX = srcData(userSrcNum)%xPixel
  bkgAnnulus(1)%centreY = srcData(userSrcNum)%yPixel
  bkgAnnulus(1)%innerLongAxis = srcData(userSrcNum)%region%longSemiAxis
else
        bkgAnnulus%centreX = srcData%xPixel
        bkgAnnulus%centreY = srcData%yPixel

        bkgAnnulus%innerLongAxis = srcData%region%longSemiAxis
end if
        bkgAnnulus%outerLongAxis = 3.0 * bkgAnnulus%innerLongAxis
!*** should be some user control over both???

        bkgAnnulus%ellipse%longSemiAxis  = 1.0
        bkgAnnulus%ellipse%isCircle  = .true.
      end if

    case('detxy')
      bkgAnnulus%unit = 'detxy'

      if (annulusConcentricAboutFovCtr) then
        bkgAnnulus%centreX = fovCtrDetX
        bkgAnnulus%centreY = fovCtrDetY

if (opMode .eq. 'single') then
  allocate(dx(1), dy(1))
  allocate(meanRadii(1))
  dx(1) = srcData(userSrcNum)%xDet - fovCtrDetX
  dy(1) = srcData(userSrcNum)%yDet - fovCtrDetY

  meanRadii = sqrt(dx * dx + dy * dy)
  deallocate(dx, dy)

  bkgAnnulus(1)%outerLongAxis = meanRadii(1)&
    + srcData(userSrcNum)%region%longSemiAxis
  bkgAnnulus(1)%innerLongAxis = meanRadii(1)&
    - srcData(userSrcNum)%region%longSemiAxis

else
        allocate(dx(numFilteredRows), dy(numFilteredRows))
        allocate(meanRadii(numFilteredRows))
        dx = srcData%xDet - fovCtrDetX
        dy = srcData%yDet - fovCtrDetY

        meanRadii = sqrt(dx * dx + dy * dy)
        deallocate(dx, dy)

        bkgAnnulus%outerLongAxis = meanRadii&
          + srcData%region%longSemiAxis
        bkgAnnulus%innerLongAxis = meanRadii&
          - srcData%region%longSemiAxis
end if

        deallocate(meanRadii)

        where (bkgAnnulus%innerLongAxis <= 0.0)&
          bkgAnnulus%isAnnulus = .false.

        bkgAnnulus%ellipse%longSemiAxis  = 1.0
        bkgAnnulus%ellipse%isCircle = .true.

      else ! annulus concentric about src ctr:
if (opMode .eq. 'single') then
  bkgAnnulus(1)%centreX = srcData(userSrcNum)%xDet
  bkgAnnulus(1)%centreY = srcData(userSrcNum)%yDet
  bkgAnnulus(1)%innerLongAxis = srcData(userSrcNum)%region%longSemiAxis
else
        bkgAnnulus%centreX = srcData%xDet
        bkgAnnulus%centreY = srcData%yDet

        bkgAnnulus%innerLongAxis = srcData%region%longSemiAxis
end if
        bkgAnnulus%outerLongAxis = 3.0 * bkgAnnulus%innerLongAxis
!*** should be some user control over both???

        bkgAnnulus%ellipse%longSemiAxis  = 1.0
        bkgAnnulus%ellipse%isCircle  = .true.
      end if

    case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
    end select

    !...................................................................
    ! Construct the region specifications and save to file:

    call message(AppMsg, verbose, "Constructing and saving the &
      &region specifications.")

    select case(opMode)
    case('single')
      allocate(regionExtnData(1))

      call constructSrcRegionSpec(srcData(userSrcNum), srcShrinkage(userSrcNum)&
        , srcRegion)
!      call constructBkgRegionSpec(bkgAnnulus(userSrcNum), srcData, bkgRegions)
call constructBkgRegionSpec(bkgAnnulus(1), srcData, bkgRegions)

      regionExtnData(1)%extName = 'REGION'

      select case(outUnit)
      case('xy')
        regionExtnData(1)%xColName = 'X'
        regionExtnData(1)%yColName = 'Y'
        regionExtnData(1)%withWcs = .true.
        regionExtnData(1)%wcs = copyWcsT(wcs) ! in ssclib/wcs-aux

      case('detxy')
        regionExtnData(1)%xColName = 'DETX'
        regionExtnData(1)%yColName = 'DETY'
        regionExtnData(1)%withWcs = .false.

      case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
      end select

      allocate(regionExtnData(1)%regions(1))
      regionExtnData(1)%regions = copyRegionT(srcRegion)
      call writeRegionExtension(srcRegionSetName, regionExtnData)
      deallocate(regionExtnData(1)%regions)

      regionExtnData(1)%regions => bkgRegions
      call writeRegionExtension(bkgRegionSetName, regionExtnData)
      deallocate(bkgRegions) ! Note that this also deallocates
      ! regionExtnData(1)%regions, since this points to the same space
      ! as bkgRegions.

      set = dataSet(srcRegionSetName, MODIFY)
        call setAttribute(set, 'SRCID', userSrcNum)
      call release(set)
      set = dataSet(bkgRegionSetName, MODIFY)
        call setAttribute(set, 'SRCID', userSrcNum)
      call release(set)

    case('batch')
      allocate(regionExtnData(min(maxSrcNum, numFilteredRows)))
      allocate(bkgRegList(min(maxSrcNum, numFilteredRows)))
      allocate(srcRegions(min(maxSrcNum, numFilteredRows)))

      do srcNum = 1, min(maxSrcNum, numFilteredRows)
        call constructSrcRegionSpec(srcData(srcNum), srcShrinkage(srcNum)&
          , srcRegions(srcNum))
        call constructBkgRegionSpec(bkgAnnulus(srcNum), srcData&
          , bkgRegList(srcNum)%regions)

        srcIdStr = intToStr(int(srcData(srcNum)%idNum)&
          , leftJustify=.true._boolean) ! in ssclib/ssc_misc
        regionExtnData(srcNum)%extName = 'REG_'//srcIdStr

        select case(outUnit)
        case('xy')
          regionExtnData(srcNum)%xColName = 'X'
          regionExtnData(srcNum)%yColName = 'Y'
          regionExtnData(srcNum)%withWcs = .true.
          regionExtnData(srcNum)%wcs = copyWcsT(wcs) ! in ssclib/wcs_aux

        case('detxy')
          regionExtnData(srcNum)%xColName = 'DETX'
          regionExtnData(srcNum)%yColName = 'DETY'
          regionExtnData(srcNum)%withWcs = .false.

        case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
        end select
      end do

      ! Write the source regions to a single file:

      do srcNum = 1, min(maxSrcNum, numFilteredRows)
        allocate(regionExtnData(srcNum)%regions(1))
        regionExtnData(srcNum)%regions = copyRegionT(srcRegions(srcNum))
      end do
      call writeRegionExtension(srcRegionSetName, regionExtnData)

      deallocate(srcRegions)

      ! Now write the background regions to a single file:

      do srcNum = 1, min(maxSrcNum, numFilteredRows)
        deallocate(regionExtnData(srcNum)%regions)
        regionExtnData(srcNum)%regions => bkgRegList(srcNum)%regions
      end do
      call writeRegionExtension(bkgRegionSetName, regionExtnData)

      ! Finally, deallocate all these region pointers. Note that the
      ! following also deallocates regionExtnData(srcNum)%regions, since
      ! this points to the same space as bkgRegList(srcNum)%regions.

      do srcNum = 1, min(maxSrcNum, numFilteredRows)
        deallocate(bkgRegList(srcNum)%regions)
      end do

      ! Write keywords containing the src id column name:

      set = dataSet(srcRegionSetName, MODIFY)
        call setAttribute(set, 'SRCIDCOL', srcIdColName)
      call release(set)
      set = dataSet(bkgRegionSetName, MODIFY)
        call setAttribute(set, 'SRCIDCOL', srcIdColName)
      call release(set)

    case('global')
      allocate(regionExtnData(1))

      call constructGlobalBkgRegionSpec(srcData, bkgRegions)

      regionExtnData(1)%extName = 'REGION'

      select case(outUnit)
      case('xy')
        regionExtnData(1)%xColName = 'X'
        regionExtnData(1)%yColName = 'Y'
        regionExtnData(1)%withWcs = .true.
        regionExtnData(1)%wcs = copyWcsT(wcs) ! in ssclib/wcs_aux

      case('detxy')
        regionExtnData(1)%xColName = 'DETX'
        regionExtnData(1)%yColName = 'DETY'
        regionExtnData(1)%withWcs = .false.

      case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
      end select

      regionExtnData(1)%regions => bkgRegions
      call writeRegionExtension(bkgRegionSetName, regionExtnData)
      deallocate(bkgRegions) ! Note that this also deallocates
      ! regionExtnData(1)%regions, since this points to the same space
      ! as bkgRegions.

    case default
call error('badOperationsStyle',"The value of the --operationstyle parameter &
&is not recognized.")
    end select

    !...................................................................
    ! Save various keywords to both output sets:

!    call message(AppMsg, verbose, "Saving some keywords.")
!
!    if (opMode .ne. 'global') then
!      set = dataSet(srcRegionSetName, MODIFY)
!      call setAttribute(set, 'SEL_EXPR', selectionExpression)
!      call setAttribute(set, 'RAD_MODE', radiusMode)
!      call setAttribute(set, 'SHRUNK', logical(adjustRadius))
!
!      select case(radiusMode)
!      case('contour')
!        call setAttribute(set, 'BKG_FRAC', fracBackground)
!        call setAttribute(set, 'BKG_MODE', bkgFluxStyle)
!        call setAttribute(set, 'ELLIPTIC', logical(doEllipse))
!
!        select case(bkgFluxStyle)
!        case('col') ! stick with what we have
!          call setAttribute(set, 'BKG_COL', bkgRateColName)
!        case('user')
!          call setAttribute(set, 'BKG_USER', userBkgFluxPerPixel, 'flux/pixel')
!        case default
!call error('badBkgFluxStyle', "The value of --bkgratestyle not recognized.")
!        end select
!
!      case('userfixed')
!        call setAttribute(set, 'RAD_USER', userSrcRadiusArcsec, 'arcsec')
!      case('enfrac')
!        call setAttribute(set, 'FRAC_E', fracEncircEnergy)
!      case default
!call error('badRadiusStyle', "The value of --radiusstyle not recognized.")
!      end select
!      call release(set)
!    end if
!
!    set = dataSet(bkgRegionSetName, MODIFY)
!    call setAttribute(set, 'SEL_EXPR', selectionExpression)
!    call setAttribute(set, 'RAD_MODE', radiusMode)
!!    call setAttribute(set, 'SHRUNK', logical(adjustRadius))
!!*** this isn't currently done in global mode.
!
!    select case(radiusMode)
!    case('contour')
!      call setAttribute(set, 'BKG_FRAC', fracBackground)
!      call setAttribute(set, 'BKG_MODE', bkgFluxStyle)
!      call setAttribute(set, 'ELLIPTIC', logical(doEllipse))
!
!      select case(bkgFluxStyle)
!      case('col') ! stick with what we have
!        call setAttribute(set, 'BKG_COL', bkgRateColName)
!      case('user')
!        call setAttribute(set, 'BKG_USER', userBkgFluxPerPixel, 'flux/pixel')
!      case default
!call error('badBkgFluxStyle', "The value of --bkgratestyle not recognized.")
!      end select
!
!    case('userfixed')
!      call setAttribute(set, 'RAD_USER', userSrcRadiusArcsec, 'arcsec')
!    case('enfrac')
!      call setAttribute(set, 'FRAC_E', fracEncircEnergy)
!    case default
!call error('badRadiusStyle', "The value of --radiusstyle not recognized.")
!    end select
!    call release(set)

    !...................................................................
    ! Clean up:

    deallocate(srcData, srcShrinkage)!, srcRegions)!, bkgRegions)
    deallocate(srcIdBand, srcIdInst)

  end subroutine region

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    opMode = stringParameter("operationstyle")
    select case(opMode)
    case('single')
      userSrcNum = intParameter("usersrcid")

    case('batch')
      srcIdColName = stringParameter("srcidcol")

    case('global')
    case default
call error('badOperationsStyle',"The value of the --operationstyle parameter &
&is not recognized.")
    end select

    evlistSetName=stringParameter("eventset") ! for 'image' wcs, instrument, etc

!    tempSrcSetName=stringParameter("tempset")
tempSrcSetName='temp.ds'
!*** remove this from here, and remove the parameter.

    srcListSetTabName = stringParameter("srclisttab")
    selectionExpression = stringParameter("expression")

    srcRegionSetName=stringParameter("regionset")
    bkgRegionSetName=stringParameter("bkgregionset")

    adjustRadius=booleanParameter("shrinkconfused")

    radiusMode=stringParameter("radiusstyle")
    select case(radiusMode)
    case('userfixed')
      userSrcRadiusArcsec=realParameter("fixedradius")

    case('enfrac')
      fracEncircEnergy = realParameter("energyfraction")

    case('contour')
      bkgFluxStyle = stringParameter('bkgratestyle')
      select case(bkgFluxStyle)
      case('col')
        bkgRateColName = stringParameter('bkgratecol')

      case('kwdset')
        bkgKwdSetName = stringParameter('bkgratekwdset')

      case('image')
        bkgImgSetName = stringParameter('bkgrateimgset')

      case('user')
        userBkgFluxPerPixel = realParameter('bkgrateuser')

      case default
call error('badBkgFluxStyle', "The value of --bkgratestyle not recognized.")
      end select

      fracBackground = realParameter("bkgfraction")
      doEllipse = .not. booleanParameter("nosrcellipse")

    case default
call error('badRadiusStyle', "The value of --radiusstyle not recognized.")
    end select

    annulusConcentricAboutFovCtr = booleanParameter("fovbkgannulus")

    outUnit = lcase(stringParameter('outunit')) ! in ssclib/ssc_misc
    select case(outUnit)
    case('detxy')
    case('xy')
      retainEllipticalBkgAnnuli = .not. booleanParameter("nobkgellipse")
    case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
    end select

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine getBandEdges(set, bandLo, bandHi, infoFoundThisBand)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set
    real(double), intent(out) :: bandLo(3,0:9), bandHi(3,0:9)
    logical(bool), intent(out) :: infoFoundThisBand(3,0:9)

    character(2) :: instChars(3)
    integer :: idInst, idBand
    character(8) :: loKwdName, hiKwdName

    !-------------------------------------------------------------------
    ! Start

    instChars = (/'PN', 'M1', 'M2'/)

    infoFoundThisBand = .false. ! default
    bandLo = 0.0 ! default
    bandHi = 0.0 ! default
    do idInst = 1, 3
      do idBand = 1, 9
        write(loKwdName, '(a3, i1, a4)') instChars(idInst)//'_', idBand, '_ELO'
        write(hiKwdName, '(a3, i1, a4)') instChars(idInst)//'_', idBand, '_EHI'
        if (hasAttribute(set, loKwdName)&
        .and. hasAttribute(set, hiKwdName)) then
          infoFoundThisBand(idInst,idBand) = .true.
          bandLo(idInst,idBand) = real64Attribute(set, loKwdName)
          bandHi(idInst,idBand) = real64Attribute(set, hiKwdName)
        end if
      end do

      if (any(infoFoundThisBand(idInst,1:9))) then 
        infoFoundThisBand(idInst,0) = .true.
        bandLo(idInst,0) = bandLo(idInst&
          , smallestTrueIndex(infoFoundThisBand(idInst,1:9)))
        bandHi(idInst,0) = bandHi(idInst&
          ,  largestTrueIndex(infoFoundThisBand(idInst,1:9)))
        ! These calls are in ssclib/array_utils.
      end if
    end do

  end subroutine getBandEdges

  !---------------------------------------------------------------------
  subroutine initializeAnnulusType(annulus)
  !---------------------------------------------------------------------

    implicit none

    type(localAnnulusType), intent(out) :: annulus(:)

    !-------------------------------------------------------------------
    ! Start

    annulus%centreX = 0.0d0
    annulus%centreY = 0.0d0
    annulus%innerLongAxis = 0.0
    annulus%outerLongAxis = 0.0
    call initializeEllipseT(annulus%ellipse) ! in ssclib/ellipse
    annulus%unit = 'none'
    annulus%isAnnulus = .true. ! false means that the inner ellipse is of
    ! zero size.

  end subroutine initializeAnnulusType

  !---------------------------------------------------------------------
  subroutine getEncircEnergyRadii(srcThetaArcsec, srcPhi, srcEeV&
  , fracEncircEnergy, srcRadiusTan)
  !---------------------------------------------------------------------
  ! NOTE! Assumes (i) that cal has been properly set up for the instrument.

    implicit none

    real(double), intent(in)  :: srcPhi(:),&
                                 srcThetaArcsec(size(srcPhi)),&
                                 srcEeV(size(srcPhi))
    real(single), intent(in)  :: fracEncircEnergy
    real(single), intent(out) :: srcRadiusTan(size(srcPhi))

    real(single) ::&
      srcRadiusMm(size(srcPhi))
    integer(int16) ::&
      numSources, i
    type(PsfT) ::&
      psf
    real(double) ::&
      focalLengthMm
    type(PsfValidityRangesT) :: validPsfRanges
    real(double) :: effectiveSrcThetaArcsec

    !-------------------------------------------------------------------
    ! Start

    numSources = size(srcPhi)

    validPsfRanges = CAL_psfValidityRanges()

    call CAL_setState(accuracyLevel=ACCURACY_MEDIUM)
    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)

    do i = 1, numSources
      if  (srcEeV(i) < validPsfRanges%eMin&
      .or. srcEeV(i) > validPsfRanges%eMax&
      .or. srcThetaArcsec(i) < validPsfRanges%thetaMin&
      .or. srcPhi(i) < validPsfRanges%phiMin&
      .or. srcPhi(i) > validPsfRanges%phiMax)&
call error('energyThetaOrPhiOutOfRange',"One or more of energy, theta or &
&phi were outside allowed ranges for the PSF.")

      if (srcThetaArcsec(i) > validPsfRanges%thetaMax) then
call warning('srcThetaTooLarge',"Theta for the source was outside the &
&allowed range for the PSF. The maximum valid value will be substituted.")

        effectiveSrcThetaArcsec = validPsfRanges%thetaMax&
          * (1.0d0 - epsilon(srcThetaArcsec(i)))
      else
        effectiveSrcThetaArcsec = srcThetaArcsec(i)
      end if

      psf = CAL_getPSF(srcEeV(i), effectiveSrcThetaArcsec, srcPhi(i))
      srcRadiusMm(i) = CAL_psfEncircledEnergy(psf&
        , real(fracEncircEnergy, double))
    end do

    srcRadiusTan = srcRadiusMm / focalLengthMm

  end subroutine getEncircEnergyRadii

  !---------------------------------------------------------------------
  subroutine correctRadii(srcX, srcY, srcRadius, srcShrinkage)
  !---------------------------------------------------------------------

    implicit none

    real(single),  dimension(:), intent(in) ::&
      srcX, srcY
    real(single),  dimension(:), intent(in) ::&
      srcRadius
    real(single),  dimension(:), intent(out) ::&
      srcShrinkage

    integer(int16) ::&
      numSources, i, j, confSetNum(size(srcX)), maxNumConfSets, numConfSrcs
    real(single),  dimension(size(srcX)) ::&
      tempSrcX, tempSrcY, tempSrcRadius, tempSrcRadiusShrinkage
logical :: test

    !-------------------------------------------------------------------
    ! Start

    srcShrinkage = 1.0 ! vector op, default.
  if (.not. adjustRadius) return

    numSources = size(srcX)

    call findConfusedSets(real(srcX, double), real(srcY, double), real(srcRadius, double), confSetNum)
    ! in ssclib/confusion

    maxNumConfSets = maxval(confSetNum)
  if (maxNumConfSets .le. 0) return

    do i = 1, maxNumConfSets
    if (count(confSetNum .eq. i) .le. 0) cycle

      numConfSrcs = 0
      do j = 1, numSources
        if (confSetNum(j) .eq. i .and. srcRadius(j) .gt. 0.0) then
          numConfSrcs = numConfSrcs + 1
          tempSrcX(numConfSrcs)               = srcX(j)
          tempSrcY(numConfSrcs)               = srcY(j)
          tempSrcRadius(numConfSrcs)          = srcRadius(j)
          tempSrcRadiusShrinkage(numConfSrcs) = srcShrinkage(j)
        end if
      end do

test = .false.
      call correctConfSet(tempSrcX(1:numConfSrcs), tempSrcY(1:numConfSrcs)&
        , tempSrcRadius(1:numConfSrcs)&
        , tempSrcRadiusShrinkage(1:numConfSrcs),test)

      numConfSrcs = 0
      do j = 1, numSources
        if (confSetNum(j) .eq. i .and. srcRadius(j) .gt. 0.0) then
          numConfSrcs = numConfSrcs + 1
          srcShrinkage(j) = tempSrcRadiusShrinkage(numConfSrcs)
        end if
      end do
    end do

  end subroutine correctRadii

  !---------------------------------------------------------------------
  subroutine correctConfSet(srcX, srcY, srcRadius, srcShrinkage, test)
  !---------------------------------------------------------------------

    implicit none

    real(single),  dimension(:), intent(in) ::&
      srcX,&
      srcY,&
      srcRadius
    real(single),  dimension(:), intent(inout) ::&
      srcShrinkage
logical, intent(in), optional :: test

    real(single),  dimension(size(srcX), size(srcX)) ::&
      separations,&
      overlapFraction
    real(single),  dimension(size(srcX)) ::&
      srcRadiusEffective
    integer(int16) ::&
      numSources,&
      i,&
      j,&
      iOfMinMax,&
      jOfMinMax
    real(single) ::&
      sumOfRadii,&
      dX,&
      dY,&
      overlap,&
      reduction, inflation
    logical :: noNegativeOverlaps, noPositiveOverlaps, localTest

    !-------------------------------------------------------------------
    ! Start

if (present(test)) then
  localTest = test
else
  localTest = .false.
end if

    srcRadiusEffective = srcRadius * srcShrinkage
if (minval(srcRadiusEffective) <= 0.0) call error('nonRealSrcRadii'&
, "The effective or post-shrinkage source radius is <= 0.")

    numSources = size(srcX)

    ! Do some preliminary setup:
    separations = 0.0 ! default (remains at this value on the diagonal)
    overlapFraction = 0.0 ! default (remains at this value on the diagonal)
    do i = 1, numSources - 1
      do j = i + 1, numSources
        dX = srcX(i) - srcX(j)
        dY = srcY(i) - srcY(j)
        separations(i, j) = sqrt(dX * dX + dY * dY)
        separations(j, i) = separations(i, j)
        if (separations(i, j) .le. 0.0) then
          write(errStr, '(a, i5, a, i5, a, es12.5, a, es12.5, a)') 'Sources '&
          , i,' and ', j,' have the same position, namely (x,y) = ('&
          , srcX(i),',', srcY(i),'). You might want to filter the source &
          &list to remove sources with duplicate rows.'
call error('doubledSource')
        end if

        sumOfRadii = srcRadiusEffective(i) + srcRadiusEffective(j)
        overlap = sumOfRadii - separations(i, j)
        overlapFraction(i, j) = overlap / srcRadiusEffective(i)
        overlapFraction(j, i) = overlap / srcRadiusEffective(j)

        ! Overlap: (i,j)<0 means that source i and j are not confused;
        ! 0<(i,j)<1 means that i and j are confused but centre of i is
        ! outside radius of j (not necessarily the other way around);
        ! (i,j)>1 means that i and j are confused and centre of i is
        ! inside the radius of j (but not necessarily the other way around).
      end do
    end do

    ! Now selectively shrink sources in an attempt to remove all the overlaps:
    do while(maxval(overlapFraction) .gt. 0.0)
      call findMinPosMaxOverlap(overlapFraction, noPositiveOverlaps&
      , iOfMinMax, jOfMinMax)

    if (noPositiveOverlaps) exit

      ! Eliminate the overlap between sources iOfMinMax and jOfMinMax by
      ! shrinking them by an equal multiplicative factor (reduction):

      reduction = separations(iOfMinMax, jOfMinMax)&
      / (srcRadiusEffective(iOfMinMax) + srcRadiusEffective(jOfMinMax))

      reduction = min(max(reduction, 0.0), 1.0) ! just to make sure there are
      ! no funny values.

      srcRadiusEffective(iOfMinMax) = srcRadiusEffective(iOfMinMax) * reduction
      srcRadiusEffective(jOfMinMax) = srcRadiusEffective(jOfMinMax) * reduction
      call adjustOverlaps(srcRadiusEffective, separations, overlapFraction&
        , iOfMinMax)

      call adjustOverlaps(srcRadiusEffective, separations, overlapFraction&
        , jOfMinMax)

      overlapFraction(iOfMinMax, jOfMinMax) = 0.0
      overlapFraction(jOfMinMax, iOfMinMax) = 0.0

      ! Sources other than the pair just 'shrunk' can get 'left behind',
      ! that is, they can end up with radii that are unecessarily small.
      ! To prevent this, go through the overlapFractions and see if there
      ! any with negative overlaps (indicating they are no longer confused);
      ! if so, choose that which has been shrunk the most and deshrink it.
      ! Repeat the exercise until there are no more sources which can be
      ! deshrunk.

      do
        call findMaxNegOverlap(overlapFraction, srcRadiusEffective/srcRadius&
        , noNegativeOverlaps, iOfMinMax, jOfMinMax)

if (localTest) print *, 'iOfMinMax, jOfMinMax:', iOfMinMax, jOfMinMax
if (localTest) print *, 'overlapFraction(iOfMinMax, jOfMinMax):', overlapFraction(iOfMinMax, jOfMinMax)

      if (noNegativeOverlaps) exit

        inflation = (separations(iOfMinMax, jOfMinMax)&
        - srcRadiusEffective(jOfMinMax)) / srcRadiusEffective(iOfMinMax)

        inflation = max(inflation, 1.0) ! just to make sure there are
        ! no funny values.
if (localTest) print *, 'inflation:', inflation

        srcRadiusEffective(iOfMinMax) = min(srcRadius(iOfMinMax)&
        , srcRadiusEffective(iOfMinMax) * inflation)
if (localTest) print *, 'srcRadius(iOfMinMax):', srcRadius(iOfMinMax)
if (localTest) print *, 'srcRadiusEffective(iOfMinMax) * inflation:', srcRadiusEffective(iOfMinMax) * inflation
if (localTest) print *, 'srcRadiusEffective(iOfMinMax):', srcRadiusEffective(iOfMinMax)

        call adjustOverlaps(srcRadiusEffective, separations, overlapFraction&
        , iOfMinMax)
      end do ! deshrinking loop.
    end do

    srcShrinkage = srcRadiusEffective / srcRadius

  end subroutine correctConfSet

  !---------------------------------------------------------------------
  subroutine findMinPosMaxOverlap(overlapFraction, noPositiveOverlaps&
  , iOfMinMax, jOfMinMax)
  !---------------------------------------------------------------------

    implicit none

    real(single),   intent(in)  :: overlapFraction(:,:)
    integer(int16), intent(out) :: iOfMinMax, jOfMinMax
    logical, intent(out) :: noPositiveOverlaps

    integer(int16) :: i, numSources, locVector(1)
    real(single) ::&
      maxOverlapVector(size(overlapFraction, 1)),&
      minMaxOverlapFraction,&
tempVector(size(overlapFraction, 1))

    !-------------------------------------------------------------------
    ! Start

    numSources = size(overlapFraction, 1)

    noPositiveOverlaps = .true. ! default

    ! See if there are any sources with a positive maximum overlap: 
    do i = 1, numSources

tempVector = overlapFraction(i, :)
maxOverlapVector(i) = maxval(tempVector)

!      maxOverlapVector(i) = maxval(overlapFraction(i, 1:numSources))

      ! This now contains the maximum amount of overlap suffered by each
      ! source, except that this value will be zero for sources which actually
      ! have a negative maximum overlap (ie are not confused with any other),
      ! because overlapFraction always has zero values on the diagonal.
      ! However this does not matter here since we are only in this subroutine
      ! interested in the positive values.
    end do

    if (any(maxOverlapVector .gt. 0.0)) then
      noPositiveOverlaps = .false.

      ! Find the minimum positive value of maxOverlapVector:

      minMaxOverlapFraction = maxval(maxOverlapVector) ! starting value.
locVector = maxloc(maxOverlapVector)
iOfMinMax = locVector(1)
      do i = 1, numSources
        if (maxOverlapVector(i) .gt. 0.0&
        .and. maxOverlapVector(i) .lt. minMaxOverlapFraction) then
          iOfMinMax = i
          minMaxOverlapFraction = maxOverlapVector(i)
        end if
      end do

      ! Have now identified the source (iOfMinMax) with the lowest positive
      ! value of maxOverlapVector; now find which of the other sources
      ! overlaps iOfMinMax by the maximum amount:

tempVector = overlapFraction(iOfMinMax, :)
locVector = maxloc(tempVector)
!      locVector = maxloc(overlapFraction(iOfMinMax, :))
      jOfMinMax = locVector(1)
    end if

  end subroutine findMinPosMaxOverlap

  !---------------------------------------------------------------------
  subroutine findMaxNegOverlap(overlapFraction, srcShrinkage&
  , noNegativeOverlaps, iOfMinMax, jOfMinMax)
  !---------------------------------------------------------------------

    implicit none

    real(single),   intent(in)  :: overlapFraction(:,:), srcShrinkage(:)
    integer(int16), intent(out) :: iOfMinMax, jOfMinMax
    logical, intent(out) :: noNegativeOverlaps

    integer(int16) :: i, numSources,&
      locVector(1)
    real(single) ::&
      tempVector(size(srcShrinkage)),&
      maxOverlapVector(size(srcShrinkage)),&
      shrinkage(size(srcShrinkage))
real(single) :: myPrecision

    !-------------------------------------------------------------------
    ! Start
myPrecision = 1e-6
!******

    numSources = size(srcShrinkage)

    noNegativeOverlaps = .true. ! default
    ! See if there are any sources with shrinkage and also a negative
    ! maximum overlap: 
    do i = 1, numSources
      tempVector = overlapFraction(i, :)
      tempVector(i) = minval(tempVector) ! to get rid of the meaningless
      ! zeroes on the diagonal of overlapFraction.

      maxOverlapVector(i) = maxval(tempVector)
      ! This now contains the maximum amount of overlap suffered by each
      ! source, including those sources with negative maximum overlap.
    end do

!    if (any(maxOverlapVector .lt. 0.0 .and. srcShrinkage .lt. 1.0)) then
if (any(maxOverlapVector .lt. -myPrecision .and. srcShrinkage .lt. 1.0 - myPrecision)) then
      noNegativeOverlaps = .false.

      ! For those sources which have a negative value of maxOverlapVector
      ! (ie are not now confused with any other), find that one which has
      ! suffered the maximum shrinkage:

      shrinkage = 1.0
!      where(maxOverlapVector .lt. 0.0)&
where(maxOverlapVector .lt. -myPrecision)&
        shrinkage = srcShrinkage

      locVector = minloc(shrinkage)
      iOfMinMax = locVector(1)

      ! Now for this source (iOfMinMax), find which other source is nearest
      ! to it. The gap between their radii will constrain by how much source
      ! iOfMinMax can be reinflated.

      tempVector = overlapFraction(iOfMinMax, :)
      tempVector(iOfMinMax) = minval(tempVector) ! to get rid of the meaningless
      ! zeroes on the diagonal of overlapFraction.
      locVector = maxloc(tempVector)
      jOfMinMax = locVector(1)
    end if

  end subroutine findMaxNegOverlap

  !---------------------------------------------------------------------
  subroutine adjustOverlaps(srcRadiusEffective, separations, overlapFraction&
  , iOfMinMax)
  !---------------------------------------------------------------------

    implicit none

    real(single),   intent(in)    :: srcRadiusEffective(:), separations(:,:)
    real(single),   intent(inout) :: overlapFraction(:,:)
    integer(int16), intent(in)    :: iOfMinMax

    integer(int16) :: i, numSources
    real(single)   :: sumOfRadii, overlap

    !-------------------------------------------------------------------
    ! Start

    numSources = size(srcRadiusEffective)

    do i = 1, numSources
    if (i .eq. iOfMinMax) cycle
      sumOfRadii = srcRadiusEffective(i) + srcRadiusEffective(iOfMinMax)
      overlap = sumOfRadii - separations(i, iOfMinMax)
      overlapFraction(i, iOfMinMax) = overlap / srcRadiusEffective(i)
      overlapFraction(iOfMinMax, i) = overlap / srcRadiusEffective(iOfMinMax)
    end do

  end subroutine adjustOverlaps

  !---------------------------------------------------------------------
  subroutine constructSrcRegionSpec(srcData, srcShrinkage, srcRegion)
  !---------------------------------------------------------------------

    implicit none

    type(SrcDataT), intent(in) :: srcData
    real(single), intent(in) :: srcShrinkage
    type(ims_RegionT), intent(out) :: srcRegion

    !-------------------------------------------------------------------
    ! Start

    call initializeRegionT(srcRegion) ! in ims_ascregion

    select case(outUnit)
    case('xy')
      srcRegion%x(1) = srcData%xPixel
      srcRegion%y(1) = srcData%yPixel

    case('detxy')
      srcRegion%x(1) = srcData%xDet
      srcRegion%y(1) = srcData%yDet

    case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
    end select

    srcRegion%r(1) = srcData%region%longSemiAxis * srcShrinkage
    srcRegion%component = 1

    if (srcData%region%isCircle) then
      srcRegion%shape = 'CIRCLE'

    else
      srcRegion%shape = 'ELLIPSE'
      srcRegion%r(2) = srcData%region%shortSemiAxis * srcShrinkage
      srcRegion%a(1) = srcData%region%rotatedAngle / PiOn180 + 90.0
    end if

  end subroutine constructSrcRegionSpec

  !---------------------------------------------------------------------
  subroutine constructBkgRegionSpec(bkgAnnulus, srcData, bkgRegions)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate the pointer bkgRegions
  ! after use.

    implicit none

    type(localAnnulusType), intent(in) :: bkgAnnulus
    type(SrcDataT), intent(in) :: srcData(:)
    type(ims_RegionT), pointer :: bkgRegions(:)

    integer :: numSrcs, srcNum, numConfusedSources, confSrcNum
    real(single) :: dX(size(srcData)), dY(size(srcData)), srcRadii(size(srcData)), minRadius, maxRadius
    logical(bool) :: srcConfusedWithBkg(size(srcData))
    type(ims_RegionT) :: confSrcRegion

    !-------------------------------------------------------------------
    ! Start

    !...................................................................
    ! First, calculate how many sources overlay the background region.
    ! This number + 1 will be the size of the bkgRegions pointer.

    numSrcs = size(srcData)

    if (bkgAnnulus%ellipse%isCircle) then ! it is relatively easy to
    ! determine those sources which are confused (or nearly confused)
    ! with the background annulus.

      select case(outUnit)
      case('xy')
        dX = srcData%xPixel - bkgAnnulus%centreX
        dY = srcData%yPixel - bkgAnnulus%centreY

      case('detxy')
        dX = srcData%xDet - bkgAnnulus%centreX
        dY = srcData%yDet - bkgAnnulus%centreY

      case default
call error('badOutUnit',"The value of the --outunit parameter is invalid.")
      end select

      srcRadii = dX * dX + dY * dY

      srcConfusedWithBkg = .false. ! default
      do srcNum = 1, numSrcs
        minRadius = bkgAnnulus%innerLongAxis&
          - srcData(srcNum)%region%longSemiAxis
        maxRadius = bkgAnnulus%outerLongAxis&
          + srcData(srcNum)%region%longSemiAxis

        if (  srcRadii(srcNum) >= minRadius * minRadius&
        .and. srcRadii(srcNum) <= maxRadius * maxRadius) then
        ! Treats all source regions as circles, but that's ok, no harm
        ! in being conservative.

          srcConfusedWithBkg(srcNum) = .true.
        end if
      end do

    else ! bkg annulus is elliptical; forget trying to tell which sources
    ! are excluded or not, this is a messy geometrical problem. Simply
    ! exclude all the sources.

      srcConfusedWithBkg = .true.
    end if

    !...................................................................
    ! Now construct the background annulus:

    numConfusedSources = count(srcConfusedWithBkg)
    allocate(bkgRegions(numConfusedSources + 1))
    call initializeRegionT(bkgRegions) ! in ims_ascregion

    if (bkgAnnulus%ellipse%isCircle) then
      if (bkgAnnulus%isAnnulus) then
        bkgRegions(1)%shape = 'ANNULUS'
        bkgRegions(1)%r(1) = bkgAnnulus%innerLongAxis
        bkgRegions(1)%r(2) = bkgAnnulus%outerLongAxis

      else ! no inner ring
        bkgRegions(1)%shape = 'CIRCLE'
        bkgRegions(1)%r(1) = bkgAnnulus%outerLongAxis
      end if

    else ! elliptical
      if (bkgAnnulus%isAnnulus) then
        bkgRegions(1)%shape = 'ELLIPTANNULUS'

        bkgRegions(1)%r(1) = bkgAnnulus%innerLongAxis
        bkgRegions(1)%r(2) = bkgAnnulus%innerLongAxis&
          * bkgAnnulus%ellipse%shortSemiAxis
        bkgRegions(1)%r(3) = bkgAnnulus%outerLongAxis
        bkgRegions(1)%r(4) = bkgAnnulus%outerLongAxis&
          * bkgAnnulus%ellipse%shortSemiAxis
        bkgRegions(1)%a(1) = bkgAnnulus%ellipse%rotatedAngle / PiOn180 + 90.0
        bkgRegions(1)%a(2) = bkgAnnulus%ellipse%rotatedAngle / PiOn180 + 90.0

      else ! no inner ring
        bkgRegions(1)%shape = 'ELLIPSE'

        bkgRegions(1)%r(1) = bkgAnnulus%outerLongAxis
        bkgRegions(1)%r(2) = bkgAnnulus%outerLongAxis&
          * bkgAnnulus%ellipse%shortSemiAxis
        bkgRegions(1)%a(1) = bkgAnnulus%ellipse%rotatedAngle / PiOn180 + 90.0
      end if
    end if

    bkgRegions(1)%x(1) = bkgAnnulus%centreX
    bkgRegions(1)%y(1) = bkgAnnulus%centreY
    bkgRegions(1)%component = 1

    !...................................................................
    ! Now add the excluded (because confused with bkg region) sources:

    confSrcNum = 0
    do srcNum = 1, numSrcs
    if (.not. srcConfusedWithBkg(srcNum)) cycle
      confSrcNum = confSrcNum + 1
      call constructSrcRegionSpec(srcData(srcNum), 1.0, confSrcRegion)
      confSrcRegion%shape = '!'//confSrcRegion%shape
      bkgRegions(confSrcNum + 1) = copyRegionT(confSrcRegion) ! in ims_ascregion
    end do

  end subroutine constructBkgRegionSpec

  !---------------------------------------------------------------------
  subroutine constructGlobalBkgRegionSpec(srcData, bkgRegions)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate the pointer bkgRegions
  ! after use.

    implicit none

    type(SrcDataT), intent(in) :: srcData(:)
    type(ims_RegionT), pointer :: bkgRegions(:)

    integer :: numSrcs, srcNum
    type(ims_RegionT) :: confSrcRegion

    !-------------------------------------------------------------------
    ! Start

    numSrcs = size(srcData)
    allocate(bkgRegions(numSrcs))
    do srcNum = 1, numSrcs
      call constructSrcRegionSpec(srcData(srcNum), 1.0, confSrcRegion)
      confSrcRegion%shape = '!'//confSrcRegion%shape
      bkgRegions(srcNum) = copyRegionT(confSrcRegion) ! in ims_ascregion
    end do

  end subroutine constructGlobalBkgRegionSpec

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module region_mod
